/********************************************************************** 
 Freeciv - Copyright (C) 2005 - The Freeciv Project
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

#include "fcintl.h"
#include "log.h"
#include "shared.h"
#include "support.h"

#include "game.h"
#include "government.h"
#include "map.h"
#include "movement.h"
#include "terrain.h"

#include "citytools.h"
#include "cityturn.h"
#include "plrhand.h"
#include "unittools.h"
#include "hand_gen.h"
#include "maphand.h"

static bool edit_mode = FALSE;

/****************************************************************************
  Handles new tile information from the client, to make local edits to
  the map.
****************************************************************************/
void handle_edit_mode(struct connection *pc, bool state)
{
  if (edit_mode == FALSE && state == TRUE) {
    /* Someone could be cheating! Warn people. */
    notify_player(NULL, NULL, E_SETTING, _(" *** Server set to edit mode! "
                  "*** "));
  }
  edit_mode = state;
}

/****************************************************************************
  Handles new tile information from the client, to make local edits to
  the map.
****************************************************************************/
void handle_edit_tile(struct connection *pc, int x, int y,
                      Terrain_type_id terrain, bv_special special)
{
  struct tile *ptile = map_pos_to_tile(x, y);
  struct terrain *pterrain = get_terrain(terrain), *old_terrain;

  if (pc->access_level != ALLOW_HACK
      || !ptile || !pterrain || !edit_mode) {
    return;
  }

  old_terrain = ptile->terrain;

  specials_iterate(s) {
    if (contains_special(special, s) && !tile_has_special(ptile, s)) {
      tile_add_special(ptile, s);
    } else if (!contains_special(special, s) && tile_has_special(ptile, s)) {
      tile_remove_special(ptile, s);
    }
  } specials_iterate_end;
  tile_change_terrain(ptile, pterrain);

  /* Handle global side effects. */
  check_terrain_change(ptile, old_terrain);

  /* update playertiles and send updates to the clients */
  update_tile_knowledge(ptile);
  send_tile_info(NULL, ptile);
}

/****************************************************************************
  We do some checking of input but not enough.
****************************************************************************/
void handle_edit_unit(struct connection *pc, struct packet_edit_unit *packet)
{
  struct tile *ptile = map_pos_to_tile(packet->x, packet->y);
  struct unit_type *punittype = get_unit_type(packet->type);
  struct player *pplayer = get_player(packet->owner);
  struct unit *punit;

  if (pc->access_level != ALLOW_HACK
      || !ptile || !punittype || !pplayer || !edit_mode) {
    return;
  }

  /* check if a unit with this id already exists on this tile, if
   * so, then this is an edit, otherwise, we create a new unit */
  if (packet->create_new) {
    struct city *homecity
      = player_find_city_by_id(pplayer, packet->homecity);

    if (is_non_allied_unit_tile(ptile, pplayer)) {
      notify_player(pplayer, ptile, E_NOEVENT,
                    _("Cannot create unit on enemy tile."));
      return;
    }
    /* FIXME: should use can_unit_exist_at_tile here. */
    if (!(ptile->city
	  && !(is_sailing_unittype(punittype)
	       && !is_ocean_near_tile(ptile)))
	&& !is_native_terrain(punittype, ptile->terrain)) {
      notify_player(pplayer, ptile, E_NOEVENT,
                    _("Cannot create %s unit on this terrain."),
                    punittype->name);
      return;
    }

    punit = create_unit(pplayer, ptile, punittype,
                        packet->veteran,
			homecity ? homecity->id : 0,
			packet->movesleft);
  } else {
    punit = find_unit_by_id(packet->id);
    if (!punit) {
      freelog(LOG_ERROR, "can't find unit to edit!");
      return;
    } 
  }

  if (packet->delete) {
    wipe_unit(punit);
    return;
  }

  punit->hp = CLIP(0, packet->hp, punittype->hp);
  punit->activity_count = packet->activity_count;
  punit->fuel = packet->fuel;
  punit->paradropped = packet->paradropped;
  if (find_unit_by_id(packet->transported_by)) {
    punit->transported_by = packet->transported_by;
  } else {
    punit->transported_by = -1;
  }

  /* FIXME: resolve_unit_stacks? */
  /* FIXME: refresh homecity? */

  /* update playertiles and send updates to the clients */
  update_tile_knowledge(ptile);
  send_unit_info(NULL, punit);
}

/****************************************************************************
  We do some checking of input but not enough.
****************************************************************************/
void handle_edit_city(struct connection *pc, struct packet_edit_city *packet)
{
  struct tile *ptile = map_pos_to_tile(packet->x, packet->y);
  struct city *pcity;
  struct player *pplayer = get_player(packet->owner);
  int i;
  int old_traderoutes[NUM_TRADEROUTES];

  if (pc->access_level != ALLOW_HACK
      || !edit_mode
      || !pplayer
      || pplayer->is_observer
      || !ptile) {
    return;
  }

  pcity = tile_get_city(ptile);
  if (!pcity) {
    if (!city_can_be_built_here(ptile, NULL)) {
      notify_player(pplayer, ptile, E_NOEVENT,
                    _("Cannot build city on this tile."));
      return;
    }

    /* new city */
    create_city(pplayer, ptile, city_name_suggestion(pplayer, ptile));
    pcity = tile_get_city(ptile);

    if (!pcity) {
      notify_player(pplayer, ptile, E_NOEVENT, _("Could not create city."));
      return;
    }
  }

  city_change_size(pcity, CLIP(0, packet->size, MAX_CITY_SIZE));

  /* FIXME: should probably be a different packet_edit_trade_route. */
  for (i = 0; i < NUM_TRADEROUTES; i++) {
    old_traderoutes[i] = pcity->trade[i];
  }
  for (i = 0; i < NUM_TRADEROUTES; i++) {
    struct city *oldcity = find_city_by_id(old_traderoutes[i]);
    struct city *newcity = find_city_by_id(packet->trade[i]);

    if (oldcity && !newcity) {
      remove_trade_route(pcity, oldcity);
    } else if (newcity && !oldcity && can_cities_trade(pcity, newcity)) {
      establish_trade_route(pcity, newcity);
    }
  }

  pcity->food_stock = MAX(packet->food_stock, 0);
  pcity->shield_stock = MAX(packet->shield_stock, 0);

  pcity->did_buy = packet->did_buy;
  pcity->did_sell = packet->did_sell;
  pcity->was_happy = packet->was_happy;
  pcity->airlift = packet->airlift;

  /* Need better checking of these values (esp. changed_from). */
  pcity->turn_last_built = CLIP(0, packet->turn_last_built, game.info.turn);
  pcity->turn_founded = CLIP(0, packet->turn_founded, game.info.turn);
  pcity->changed_from.is_unit = packet->changed_from_is_unit;
  pcity->changed_from.value = packet->changed_from_id;
  pcity->before_change_shields = packet->before_change_shields;
  pcity->disbanded_shields = MAX(packet->disbanded_shields, 0);
  pcity->caravan_shields = MAX(packet->caravan_shields, 0);
  pcity->last_turns_shield_surplus
    = MAX(packet->last_turns_shield_surplus, 0);

  /* make everything sane.  Note some refreshes may already have been
   * done above. */
  city_refresh(pcity);

  /* send update back to client */
  send_city_info(NULL, pcity);  
}

/**************************************************************************
 right now there are no checks whatsoever in the server. beware.
***************************************************************************/
void handle_edit_player(struct connection *pc, 
                        struct packet_edit_player *packet)
{
  struct player *pplayer = get_player(packet->playerno);
  struct player_research *research;
  int i;

  if (pc->access_level != ALLOW_HACK
      || !edit_mode
      || !pplayer) {
    return;
  }

  sz_strlcpy(pplayer->name, packet->name);
  sz_strlcpy(pplayer->username, packet->username);  

  pplayer->is_observer = packet->is_observer;
  pplayer->nation = get_nation_by_idx(packet->nation);
  pplayer->is_male = packet->is_male;
  pplayer->team = team_get_by_id(packet->team);

  pplayer->economic.gold = packet->gold;
  pplayer->economic.tax = packet->tax;
  pplayer->economic.science = packet->science;
  pplayer->economic.luxury = packet->luxury;
  pplayer->government = get_government(packet->government);
  pplayer->target_government = get_government(packet->target_government);
  pplayer->embassy = packet->embassy;
  pplayer->gives_shared_vision = packet->gives_shared_vision;
  pplayer->city_style = packet->city_style;
  for (i = 0; i < MAX_NUM_PLAYERS + MAX_NUM_BARBARIANS; i++) {
    pplayer->ai.love[i] = packet->love[i];
  }
  
  for (i = 0; i < MAX_NUM_PLAYERS + MAX_NUM_BARBARIANS; i++) {
    pplayer->diplstates[i].type = packet->diplstates[i].type;
    pplayer->diplstates[i].turns_left = packet->diplstates[i].turns_left;
    pplayer->diplstates[i].contact_turns_left
      = packet->diplstates[i].contact_turns_left;
    pplayer->diplstates[i].has_reason_to_cancel
      = packet->diplstates[i].has_reason_to_cancel;
  }

  for (i = 0; i < B_LAST/*game.num_impr_types*/; i++) {
     pplayer->small_wonders[i] = packet->small_wonders[i];
  }
  
  pplayer->ai.science_cost = packet->science_cost;

  pplayer->bulbs_last_turn = packet->bulbs_last_turn;

  research = get_player_research(pplayer);
  research->bulbs_researched = packet->bulbs_researched;
  research->techs_researched = packet->techs_researched;
  research->researching = packet->researching;
  research->future_tech = packet->future_tech;
  research->tech_goal = packet->tech_goal;

  pplayer->is_alive = packet->is_alive;
  pplayer->ai.barbarian_type = packet->barbarian_type;
  pplayer->revolution_finishes = packet->revolution_finishes;
  pplayer->ai.control = packet->ai;

  /* FIXME: and probably a bunch more stuff here */

  /* send update back to client */
  send_player_info(NULL, pplayer);  
}
